// Savage.c - General System Functions
// ---------------------------------------------------
// Copyright (C)Savage Development.
// By Matthew Tighe
// ***************************************************
// ***************************************************

#include "gba.h"
#include "savage.h"
#include <malloc.h>

// Initialise the VRAM heap, 128k at 0x60000000
VRAMBLOCK *vramHeap = NULL;
VRAMBLOCK *spriteHeap = NULL;

VRAMBLOCK *vramInit(void *pBase, DWORD ddLength)
{
	vramHeap = malloc( sizeof( VRAMBLOCK ) );
	if( vramHeap == NULL )
		return( NULL );

	vramHeap->ddFlags = 0;				// Free, un-padded
	vramHeap->ddLength = ddLength;
	vramHeap->pAddress = pBase;
	vramHeap->pPrev = NULL;
	vramHeap->pNext = NULL;

	return( vramHeap );
}

// vramAlloc( ddSize, dwAlign )
// ---------------------------------------------------
// Allocates a portion of vram, start aligned to <dwAlign>

void *vramAlloc(DWORD ddSize, DWORD ddAlign, WORD dwFlags)
{
	VRAMBLOCK *pBlock;
	DWORD ddPad;

	if( dwFlags & VRAM_ALLOC_OBJ )
		pBlock = spriteHeap;
	else
		pBlock = vramHeap;

	if( pBlock == NULL )
	{
		if( dwFlags & VRAM_ALLOC_OBJ )
			pBlock = vramInit( (void *) VRAM_OBJ_BASE, VRAM_OBJ_SIZE);
		else

			pBlock = vramInit( (void *) VRAM_BASE, VRAM_SIZE);
	}
	
	// Loop through all blocks to find a free on big enough.

	do
	{
		// Sort out the padding needed for alignment.
		ddPad = CalcPadding( pBlock, ddSize, ddAlign );

		// If block is big enough, use it.
		if( ddPad < 0xFFFFFFFF )
		{
			pBlock = vramUseBlock(pBlock, ddSize, ddPad);

			return( pBlock->pAddress );	// Return address in vram
		}

		pBlock = pBlock->pNext;			// Move to the next block.

	} while( pBlock != NULL );

	return( NULL );
}

// void vramUseBlock(VRAMBLOCK *pBlock, DWORD ddSize, WORD dwPad)
// ---------------------------------------------------
// Uses up <dwSize> bytes at <dwPad> alignment of memory block
// <pBlock>.
// If necessary, creates a new VRAMBLOCK to hold spare memory.

VRAMBLOCK *vramUseBlock(VRAMBLOCK *pBlock, DWORD ddSize, DWORD ddPad)
{
	VRAMBLOCK *pNewBlock;
	DWORD ddSpare;

	// If padding is big, split the block so we don't waste VRAM
	if( ddPad > 64 )
	{
		pNewBlock = malloc( sizeof( VRAMBLOCK ) );
		if( pNewBlock == NULL )
			return( NULL );

		pNewBlock->ddFlags = 0;				// Free, un-padded
		pNewBlock->ddLength = ddPad;		// Set size of new block
		pNewBlock->pAddress = pBlock->pAddress;

		pNewBlock->pNext = pBlock;			// Insert our new block into 
		pNewBlock->pPrev = pBlock->pPrev;	// ... the linked list
		pNewBlock->pPrev->pNext = pNewBlock;

		pBlock->pPrev = pNewBlock;			// ...
		pBlock->pAddress = (pBlock->pAddress + ddPad );
		pBlock->ddLength -= ddPad;

		ddPad = 0;
	}
	
	ddSpare = pBlock->ddLength - ( ddSize + ddPad );	// Calc spare bytes

	pBlock->ddFlags = ddPad | VRAM_BLOCK_USED;	// Set pre-padding

	pBlock->ddLength = ddSize;
	pBlock->pAddress = ( pBlock->pAddress + ddPad );

	// If space left at the end is big, split the block again.
	if( ddSpare > 0 )
	{
		// Allocate the memory for VRAMBLOCK structure

		pNewBlock = malloc( sizeof( VRAMBLOCK ) );
		if( pNewBlock == NULL )
			return( NULL );

		pNewBlock->ddFlags = 0;				// Free, un-padded
		pNewBlock->ddLength = ddSpare;		// Set size of new block
		pNewBlock->pAddress = ( pBlock->pAddress + ddSize );
		pNewBlock->pPrev = pBlock;			// Insert our new block into 
		pNewBlock->pNext = pBlock->pNext;	// ... the linked list
		pBlock->pNext = pNewBlock;			// ...
	}

	return( pBlock );
}

// WORD CalcPadding(VRAMBLOCK *pBlock, DWORD ddSize, WORD dwAlign)
// ---------------------------------------------------
// Calculates the padding needed to hold <ddSize> bytes in
// block <pBlock> at <dwAlign> alignment.
// RETURNS:
// --------
// Size padding req'd or -1 if block is too small.

DWORD CalcPadding(VRAMBLOCK *pBlock, DWORD ddSize, DWORD ddAlign)
{
	DWORD ddPad;

	if( pBlock->ddFlags & VRAM_BLOCK_USED )
		return( -1 );	// Fail: block is used.
	
	// Calc difference between pAddress and alignment

	ddPad = ( DWORD )(pBlock->pAddress) & ( ddAlign - 1 );
	if( ddPad > 0 )
		ddPad = ( ddAlign - ddPad );

	// If block is too small, or used, fail.
	if( pBlock->ddLength < (ddSize + ddPad ) )
		ddPad = -1;


	return( ddPad );
}