// Sprite.c - GBA Sprite Handling Functions
// ---------------------------------------------------
// Copyright (C)Savage Development.
// By Matthew Tighe
// ***************************************************
// ***************************************************

#include "gba.h"
#include "savage.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

// Macros to avoid casting all the time
#define SPR_PREV(n) ( (GFXSPRITE *)n->pPrev )
#define SPR_NEXT(n) ( (GFXSPRITE *)n->pNext )

#define SPR_Z(n) ( (GFXSPRITE *)n->z )

// Lookup tables for GBA size field
u16 shapes[4] = {
	0,		// normal
	1,		// horz rectang
	2,		// vert rectang
	4		// prohibited
};

u16 sizes[4] = {
	0,		// 1x
	1,		// 2x
	2,		// 3x
	4		// 4x
};

// sprInit
// ---------------------------------------------------
// Initialises a graphics list, which will contain sprite
// pointers.

void sprInit( GFXLIST *pList)
{
	GFXSPRITE *pSpr;
	
	oamClear( 0, 128 );		// Clear OAM memory
	
	memset( pList, 0, sizeof( GFXLIST ) );			// Clear list memorry

	// Allocate the memory for the header sprite structure
	pSpr = malloc( sizeof( GFXSPRITE ) );
	if( pSpr == NULL )
		return;

	memset( pSpr, 0, sizeof( GFXSPRITE ) );		// Clear it

	
	pList->pHead = pSpr;						// Header sprite is next to list.

	pSpr->type = 255;							// Header sprite
	pSpr->alpha = 255;					// Not visible

	pSpr->pNext = pSpr;							// Circular links coz, list is empty
	pSpr->pPrev = pSpr;
}

// sprCreate
// ---------------------------------------------------
// Allocate memory for a new GFXSPRITE an initialises
// it.

GFXSPRITE *sprCreate(GFXLIST *pList)
{
	GFXSPRITE *pSpr,*pTemp;

	// Allocate the memory for the sprite structure
	pSpr = malloc( sizeof( GFXSPRITE ) );
	if( pSpr == NULL )
		return( NULL );

	memset( pSpr, 0, sizeof( GFXSPRITE ) );	// Clear it
	
	// Insert new sprite at end of the list
	
	pTemp = pList->pHead->pPrev;

	pSpr->pPrev = pTemp;					// Point prev member to sprite previous to header sprite (last sprite)
	pTemp->pNext = pSpr;					// Insert new sprite into the chain
	
	pSpr->pNext = pList->pHead;				// Point next member to header sprite
	pList->pHead->pPrev = pSpr;				// Point sprite previous to header to us, we are the last sprite

	pList->number++;						// Inc items in list

	return( pSpr );							// Return pointer to this sprite.
}

// sprDestroy
// ---------------------------------------------------
// Removes a sprite from the linked list and frees its
// memory.

void sprDestroy(GFXLIST *pList, GFXSPRITE *pSpr)
{

	SPR_PREV( pSpr )->pNext = pSpr->pNext;	// Remove link from list

	SPR_NEXT( pSpr )->pPrev = pSpr->pPrev;	// Remove link
	
	pList->number--;

	free( pSpr );								// Free the memory
}

// sprDraw
// ---------------------------------------------------
// 'Draws' the sprites by setting up the OAM ram.
// NOTE: This is a quick hack,
// No ZORDER
// No 16 colour
// No FLAGS
// No ASSEMBLER!!!!!
// Only Type 0

void sprDraw(GFXLIST *pList, u32 uFlags)
{
	OAMOBJECT *pOam;		// Pointer to OAM ram
	GFXSPRITE *pSpr;
	unsigned int i;
	WORD x,y;

	if( uFlags & SPRDRAW_ZORDER )
		sprSort( pList );
	
	pOam = ( OAMOBJECT * ) 0x07000000;	// Set up OAM ptr
	pSpr = (GFXSPRITE *) pList->pHead->pNext;	// Point to first sprite

	// Loop for each sprite
	for(i = 0; i < pList->number; i++)
	{
		if( pSpr->type == 255 )
			continue;				// Ignore header sprite
		
		x = pSpr->x;
		y = pSpr->y;

		if( pSpr->pAttach != NULL )
		{
			x += pSpr->pAttach[0];
			y += pSpr->pAttach[1];
		}

		pOam[i].y = y | OBJ_SHAPE( shapes[(pSpr->size >> 2)] ) | OBJ_COLOUR_256x1;
		pOam[i].x = x | OBJ_SIZE( sizes[(pSpr->size % 4)] );
		pOam[i].tile = pSpr->tile;

		pSpr = SPR_NEXT( pSpr );
	}
}

// sprAttach
// ---------------------------------------------------
// Attaches a slave sprite to a master sprite.

void sprAttach(GFXSPRITE *pSlave, GFXSPRITE *pMaster)
{
	if( pMaster == NULL )
		pSlave->pAttach = NULL;
	else
		pSlave->pAttach = &pMaster->x;
}

// oamClear( nStart, nCount )
// ---------------------------------------------------
// Clears all OAM entries to zero starting at <nStart>
// for <nCount> entries.

void oamClear( int nStart, int nCount )
{
	int i;
	OAMOBJECT *pSprites;

	pSprites = ( OAMOBJECT * ) 0x07000000;


	// NOTE:To Turn-off a sprite, we must set Double Size: true Rot Scaling: false
	
	for(i = nStart; i < nCount; i++)
	{
		pSprites[i].x = 0;
		pSprites[i].y = OBJ_SDOUBLE | OBJ_SPECIAL;
		pSprites[i].tile = 0;
		pSprites[i].fx = 0;
	}
}


// Simple bubble sort routine
// --------------------------------------------------------------------------------
// TO BE REPLACED SOON

void sprSort(GFXLIST *pList)
{
	GFXSPRITE *pSpr, *pSwap;
	int i;

	pSpr = (GFXSPRITE *) pList->pHead->pNext;	// Point to first sprite

	i = 0;
	do
	{
		pSwap = SPR_NEXT( pSpr );


		if( pSpr->z > pSwap->z )
		{
			sprSwap( pSpr, pSwap );
			i = 0;
			pSpr = (GFXSPRITE *) pList->pHead->pNext;	// Point to first sprite
		}
		else
			pSpr = pSwap;

		i++;

	} while( i < pList->number );		// Loop until all sprites are in order
}

void sprSwap( GFXSPRITE *pDest, GFXSPRITE *pSource )
{
	GFXSPRITE *pTemp, *pReg;

	pTemp = SPR_PREV( pDest );
	pReg = SPR_PREV( pSource );

	pTemp->pNext = pSource;
	pSource->pPrev = pTemp;

	pReg->pNext = pDest;
	pDest->pPrev = pReg;

		
	pTemp = SPR_NEXT( pDest );
	pReg = SPR_NEXT( pSource );

	pTemp->pPrev = pSource;
	pSource->pNext = pTemp; 
	
	pReg->pPrev = pDest;
	pDest->pNext = pReg; 
}